//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-/////////////////////////////////////////////////////////
// FieldObject.h - Objects that can be displayed on a field map
// $Id: FieldObject.h,v 1.29 2002/04/08 15:52:19 cs Exp $
// Time-stamp: <2001-06-26 12:47:55 goetz>
//////////////////////////////////////////////////////////////////////

#ifndef FIELDOBJECT_H
#define FIELDOBJECT_H

#include <qpainter.h>
#include <qcolor.h>
#include <qpen.h>

#include <string>
#include <iostream.h>

#include "mcommand.h"
#include "constants.h"
#include "mrbtypes.h"
#include "coordinates.h"
#include "Exception.h"

#if (QT_VERSION >= 200)
#define QT2
#endif

#ifdef QT2
#define QTNSPC(cmd)  Qt::cmd
#else
#define QTNSPC(cmd)      cmd
#endif

class FieldMap;

class FieldObject 
{
protected:
  Vector2d pos;
public:
  static QTNSPC(PenStyle) getStyle( const MStyle style);

  // some sort of RTTI, grouped by object groups
  enum FOType { FONone, FOBase, FOServer, FOPlayer, FOEndMarker }; 
  virtual FOType type() const = 0;

  virtual void draw( QPainter &, const FieldMap * ) const = 0;
  
  FieldObject (const Vector2d &pos) 
    : pos(pos)
  {}
  virtual ~FieldObject() {};
};

/** field background with green lawn
 */
class FieldBackground : public FieldObject
{
public:
  virtual void draw( QPainter &, const FieldMap * ) const;
  virtual FOType type() const
  { return FOBase; }

  // size gives the bounding box of the widget
  FieldBackground ()
    : FieldObject( Vector2d() )
  {}
};

class FieldBall : public FieldObject
{
public:
  virtual void draw( QPainter &, const FieldMap * ) const;
  virtual FOType type() const
  { return FOServer; }

  FieldBall ( const Vector2d &pos )
    : FieldObject(pos) 
  {}
};

class WFieldBall : public FieldBall
{
protected:
  int age;
  Vector2d vel;
  QColor color;
public:
  virtual void draw( QPainter &, const FieldMap * ) const;
  virtual FOType type() const
  { return FOPlayer; }

  WFieldBall ( const Vector2d &pos,
	       int age, const Vector2d &vel, const QColor &color)
    : FieldBall( pos ), 
    age(age), vel(vel), color(color)
  {}
};

class FieldPlayer : public FieldObject
{
protected:
  bool  kicking;
  Angle bodydir;
  Angle headdir;
  Angle viewWidth;
  int   num;
  bool  leftTeam;
  bool  isGoalie;
  bool  collision;
  int   stamina;
  int   het_type;
public:
  virtual void draw( QPainter &, const FieldMap * ) const;
  virtual FOType type() const
  { return FOServer; }

  FieldPlayer( const Vector2d &pos, const Angle &bodydir, 
	       const Angle &headdir, const Angle &viewWidth,
	       const int num, const bool leftTeam,
	       const bool kicking= false, const bool isGoalie= false,
	       const bool collision= false, const int stamina= -1,
	       const int het_type= -1)
    : FieldObject(pos), kicking(kicking), bodydir(bodydir), 
    headdir(headdir), viewWidth(viewWidth), num(num), 
    leftTeam(leftTeam), isGoalie(isGoalie), collision(collision),
    stamina(stamina), het_type(het_type)
  {}
};

/** field player as used in worldmodel display
 */
class WFieldPlayer : public FieldPlayer
{
protected:
  int age;
  Vector2d vel;
  QColor color;
public:
  virtual void draw( QPainter &, const FieldMap * ) const;
  virtual FOType type() const
  { return FOPlayer; }

  WFieldPlayer( const Vector2d &pos, const Angle &bodydir, 
		const int num, const bool leftTeam,
		const int age, const Vector2d &vel, const QColor &color,
		const bool isGoalie= false)
    : FieldPlayer( pos, bodydir, Angle(0), PI_2, 
		   num, leftTeam, false, isGoalie ),
    age(age), vel(vel), color(color)
  {}
};

// various drawcommands

class FieldCircle : public FieldObject
{
protected:
  double rad;
  bool   hasgrid;
  QColor color;
  MStyle style;
public:
  virtual void draw( QPainter &, const FieldMap * ) const;
  virtual FOType type() const
  { return FOPlayer; }

  
  FieldCircle( const Vector2d &pos, double rad, 
	       bool hasgrid, const QColor &color, const MStyle style )
    : FieldObject ( pos ), rad( rad ), hasgrid( hasgrid ),
      color(color), style(style)
  {}
};

class FieldPosition : public FieldObject
{
protected:
  QColor color;
public:
  virtual void draw( QPainter &, const FieldMap * ) const;
  virtual FOType type() const
  { return FOPlayer; }
  
  FieldPosition( const Vector2d &pos, const QColor &color )
    : FieldObject( pos ), color( color )
  {}
};

class FieldLine : public FieldObject
{
protected:
  Vector2d endPos;
  QColor   color;
  MStyle  style;
public:
  virtual void draw( QPainter &, const FieldMap *) const;
  virtual FOType type() const
  { return FOPlayer; }

  FieldLine( const Vector2d &start, const Vector2d &end,
	     const QColor &color, const MStyle style )
    : FieldObject( start ), endPos( end ), color( color ), style(style)
  {}
};

class FieldVLine : public FieldObject
{
protected:
  QColor color;
  MStyle style;
public:
  virtual void draw( QPainter &, const FieldMap *) const;
  virtual FOType type() const
  { return FOPlayer; }
  
  FieldVLine( double x, const QColor &color, const MStyle style)
    : FieldObject( Vector2d(x, -FIELD_WIDTH/2) ), color( color ),
      style(style)
  {}
};

class FieldBox : public FieldObject
{
protected:
  Vector2d other;
  QColor   color;
  MStyle  style;
public:
  virtual void draw( QPainter &, const FieldMap *) const;
  virtual FOType type() const
  { return FOPlayer; }
  
  FieldBox( const Vector2d &pos1,
	    const Vector2d &pos2,
	    const QColor &color, const MStyle style )
    : FieldObject( pos1 ), other( pos2 ), color( color ), style( style )
  {}
};

class FieldCorridor : public FieldObject
{
protected:
  Vector2d target;
  double   diam;
  Angle    alpha;
  QColor   color;
  MStyle   style;
public:
  virtual void draw( QPainter &, const FieldMap *) const;
  virtual FOType type() const
  { return FOPlayer; }
  
  FieldCorridor( const Vector2d &src,
		 const Vector2d &target, double diam, const Angle &alpha,
		 const QColor &color, const MStyle style )
    : FieldObject( src ), target( target ), 
    diam( diam ), alpha( alpha ), color( color ), style( style )
  {}
};

class FieldLabel : public FieldObject
{
protected:
  string text;
  QColor color;
public:
  virtual void draw( QPainter &, const FieldMap * ) const;
  virtual FOType type() const
  { return FOPlayer; }
  
  FieldLabel( const Vector2d &pos, const string &text,
  	      const QColor &color )
      : FieldObject( pos ), text( text ), color( color )
  {}
};

class FieldMarker : public FieldObject
{
protected:
  int time;
  int duration;
  char *comm;
public:
  virtual void draw( QPainter &, const FieldMap * ) const;
  virtual FOType type() const
  { return FOBase; }

  // makes a copy of comment (no double quotes allowed!)
  FieldMarker( const Vector2d &pos, const int time, const int duration,
	       const char *comment);
  // needed for marker window
  int operator==( const FieldMarker & );
  bool isInTime(int time)
  { return this->time <= time && this->time+duration-1 >= time; }  
  
  operator char *() const;
};

ostream & operator << (ostream &os, const FieldMarker &f);

// some markers to mark the end of timesteps
class EndMarker : public FieldObject
{
public:
  const int kind;

  virtual void draw( QPainter &, const FieldMap * ) const {}
  virtual FOType type() const
  { return FOEndMarker; }
  
  EndMarker( int kind= 0 )
    : FieldObject( Vector2d() ), kind(kind)
  {}
};

class ShowinfoType : public FieldObject
{
  typedef FieldPlayer * FieldPlayer_p;
public:
  FieldBall    *ball;
  FieldPlayer_p ourTeam[11];
  FieldPlayer_p otherTeam[11];

  virtual void draw( QPainter &, const FieldMap * ) const;
  virtual FOType type() const
  { return FOServer; }

  ShowinfoType();
  
  virtual ~ShowinfoType();
};

class FieldSnapshot : public FieldObject
{
  typedef WFieldPlayer * WFieldPlayer_p;
public:
  WFieldBall    *ball;
  WFieldPlayer_p ourTeam[11];
  WFieldPlayer_p otherTeam[11];

  virtual void draw( QPainter &, const FieldMap * ) const;
  virtual FOType type() const
  { return FOPlayer; }
  
  FieldSnapshot();
  
  virtual ~FieldSnapshot();
};

#endif
