//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "MsgStore.h"
#include <iomanip>
#include <stdio.h>

// debug output
//#define DEBUG

/* if defined, all messages are logged. Messages smaller than MSG_LENGTH
   with full content, for longer messages the length is printed
*/
//#define MSG_LENGTH 512

// cwz-build-command: make FUNSSEL

MsgStore::MsgStore(const string &name) throw (Exception)
  : file(name),
    // nothing written yet
    writeTime(-2),
    minwriteTime(-2),
    numUnstored(0),
    lastWrittenTime(-2),
    curTime(-2)
{
  for (int i = 0; i < Buffered; i++)
    messages[i].head = messages[i].tail = NULL;
  readmsg.head = readmsg.tail = NULL;
  setupLocations();
}

MsgStore::~MsgStore()
{
  // flush remaining cached messages
  while (numUnstored > 0) {
    appendToFile(minwriteTime);
    minwriteTime++;
  }
  clearList(readmsg);
}

void MsgStore::setupLocations() throw (Exception)
{
  int cur_time = -1;
  Location cur_loc;
  cur_loc.fpos   = 0;
  cur_loc.length = 0;

  cerr << "fetching time indices" << endl;
  for (int t = 0; t < file.getNumRecords(); t++) {
#ifdef DEBUG
    cout << "process index " << cur_time << endl;
#else
    if (cur_time % 10 == 0)
      cerr << "\r" << "time " << cur_time;
#endif
    string msg    = file.read(t, 16);
    int file_time =  atol(msg.c_str());
    if (file_time == cur_time) {
      cur_loc.length++;
      continue;
    }
    // new time step, should be the next
    if (file_time != cur_time + 1)
      throw Exception("logfile corrupt");
#ifdef DEBUG
    cout << "index: " << cur_loc.fpos << " len "
	 << cur_loc.length << endl;
#endif
    location.push_back(cur_loc);
    cur_time++;
    cur_loc.fpos   = t;
    cur_loc.length = 1;
  }
  if (cur_loc.length > 0) {
    location.push_back(cur_loc);
#ifdef DEBUG
    cout << "index: " << cur_loc.fpos << " len "
	 << cur_loc.length << endl;
#endif
  }
  cerr << endl;
}

void MsgStore::openWrite() throw (Exception)
{
  // reopen
  file.open("", 1);
  // read starts in current list
  curTime  = -1;
  writeTime = -1;
  minwriteTime = -1;
  lastWrittenTime = -1;
  numUnstored = 0;
  writeLocation.fpos   = 0;
  writeLocation.length = 0;
  clearList(messages[timeAddr(-1)]);
  location.erase(location.begin(), location.end());
}

bool MsgStore::insertMsg(const SoccerMessage &msg) throw (Exception)
{
  if (!file.isWritable())
    throw Exception("logfile not opened for write");
  int time = msg.time;
  if (time < minwriteTime) {
    cout << minwriteTime << " " << time << endl;
    // message for a flushed time step, ignore
    return false;
  }

  while (time > writeTime) {
    MsgList &src  = messages[timeAddr(writeTime)];
    MsgList &dst = messages[timeAddr(writeTime + 1)];
    int removed = writeTime - Buffered + 1;
    // store away minwritetime if mwt hangs more than
    // Buffered timesteps
    if (removed >= minwriteTime) {
      appendToFile(minwriteTime);
      minwriteTime++;
    }
    writeTime++;
    if (curTime == removed) {
      // oops, were reading there
      // reset read
      readFromFile(curTime, readmsg);
      curMsg    = readmsg.head;
      curMsgLoc = &readmsg.head;
    }
    clearList(dst);
    MsgItem *it;
    // copy all duration > 1 messages
    for (it = src.head; it != NULL; it = it->next) {
      if (it->msg.time + it->msg.durtn > writeTime) {
	SoccerMessage s_msg(it->msg.time, it->msg.message,
			    it->msg.sender, it->msg.recver,
			    it->msg.durtn);
	addMsg(dst, s_msg);
	numUnstored +=string(s_msg).length();
      }
    }
  }

  addMsg(messages[timeAddr(time)], msg);
  numUnstored +=string(msg).length();

  checkNumUnstored();

  return true;
}

void MsgStore::checkNumUnstored()
{
#ifdef DEBUG
  cout << "store stats before: #unsaved=" << numUnstored
       << " mW=" << minwriteTime
       << " w=" << writeTime
       << " lW=" << lastWrittenTime << endl;
#endif
  while (numUnstored > MaxUnstoredBytes) {
    // if list is empty, we cannot win anything
    // but nevertheless we must save something for each step
    // so only ignore if already written
    if ((messages[timeAddr(minwriteTime)].head == NULL) &&
	(minwriteTime == lastWrittenTime)) {
      minwriteTime++;
    } 
    appendToFile(minwriteTime);
    // lastWrittenTime und minwriteTime should not differ by more than 1
    if (minwriteTime < lastWrittenTime)
      minwriteTime = lastWrittenTime;
  }

#ifdef DEBUG
  cout << "store stats after: #unsaved=" << numUnstored
       << " mW=" << minwriteTime
       << " w=" << writeTime
       << " lW=" << lastWrittenTime << endl;
#endif
}

void MsgStore::appendToFile(int t)
  throw (Exception)
{
  // buffer for 4(int + space)
  char buf[4*16];

  MsgList &lst = messages[timeAddr(t)];

  if (t < lastWrittenTime)
    return;

#ifdef DEBUG  
  cout << "saving time " << t << endl;
#endif
  string str;
  sprintf(buf, "%d ", t);
  str += string(buf);

#ifdef DEBUG  
  int count = 0;
#endif
  MsgItem *it;
  for (it = lst.head; it != NULL; it = it->next) {
    int len = string(it->msg).length();
    sprintf(buf, "%d %d %d %d %d ", it->msg.time, it->msg.durtn,
	    it->msg.sender, it->msg.recver, len);
    str += buf;
    str += it->msg;

    // update number of bytes saved
    numUnstored -= len;

#ifdef MSG_LENGTH
    if (it->msg.message.length() < MSG_LENGTH)
      cout << "saving " << it->msg.time << " " << it->msg.message << endl;
    else
      cout << "saving " << it->msg.time << " len "
	   << it->msg.message.length() << endl;
#endif
#ifdef DEBUG
    count++;
#endif
  }

#ifdef DEBUG
  cout << "saved " << count << " messages for time " << t << endl;
  cout << "still unsaved " << numUnstored << endl;
#endif
  file.push_back(str);

  // update writeLocation.
  if (t == lastWrittenTime)
    writeLocation.length++;
  else {
#ifdef DEBUG
    cout << "pushing location " << writeLocation.fpos
	 << " " << writeLocation.length << endl;
#endif
    location.push_back(writeLocation);
    if (t != lastWrittenTime + 1)
      throw Exception("logfile writing out of order");
    writeLocation.fpos   = file.getNumRecords() - 1;
    writeLocation.length = 1;
    lastWrittenTime = t;
  }

  // now that all is saved, delete list
  // but first check that it's not the currently read time step
  // in that case, we must reset the read
  if (curTime == t) {
#ifdef DEBUG
    cout << "had to reset curMsg" << endl;
#endif
    curMsg    = NULL;
    curMsgLoc = &(messages[timeAddr(curTime)].head);
  }
  clearList(lst);
}

bool MsgStore::fetchTime(int time) throw (Exception)
{
#ifdef DEBUG
  cout << "fetch " << time << " wt " << writeTime << endl;
#endif
  int correct = true;
  if (time <= -2) {
    time    = -1;
    correct = false;
  }

  if (writeTime == -2) {
    // no data written, check if in file
    if (numSaved() == 0) {
      // absolutely no data available
      // no game running, no logfile
      curTime = -2;
      return false;
    }
    if (time >= numSaved() - 1) {
      time    = numSaved() - 2;
      correct = false;
    }
  }
  else {
    // data written, check that we do not read in advance
    if (time > writeTime) {
      time = writeTime;
      correct = false;
    }
  }
  // now we have a legal time, check if we have to load
  curTime = time;

  if ((writeTime == -2) || (time < lastWrittenTime)) {
    readFromFile(time, readmsg);
    curMsg    = readmsg.head;
    curMsgLoc = &readmsg.head;
  }
  else {
    curMsg    = messages[timeAddr(curTime)].head;
    curMsgLoc = &messages[timeAddr(curTime)].head;
  }

  return correct;
}

void MsgStore::readFromFile(int t, MsgList &lst)
  throw (Exception)
{
  clearList(lst);
#ifdef DEBUG
  cout << "reading time " << t << endl;
#endif

  // no data written back for that timestep, ignore
  if (t + 1 >= (int)location.size())
    return;

  Location loc = location[t + 1];
  for (int pos = loc.fpos; pos < loc.fpos + loc.length; pos++) {
#ifdef DEBUG
    cout << "reading " << pos << endl;
#endif
    string msg = file.read(pos);
    const char *parse = msg.data();
    const char *end   = parse + msg.length();
    char       *tmp;
    int file_time =  strtol(parse, &tmp, 10);
    if (parse == tmp)
      throw Exception("file corrupt");
    parse = tmp;

    if (t != file_time) {
      cerr << "reading time " << t << " but should be " << file_time << endl;
      throw Exception("time indices corrupt");
    }

    while (parse < end) {
      SoccerMessage con_msg;
      int scanned, len;
      int i_sender, i_recver;
      int res = sscanf(parse, "%d %d %d %d %d %n", &con_msg.time,
		       &con_msg.durtn, &i_sender, &i_recver,
		       &len, &scanned);
      if (res != 5)
	return;
      con_msg.sender = (SoccerActor)i_sender;
      con_msg.recver = (SoccerActor)i_recver;
      parse += scanned;
      con_msg.message.assign(parse, len);
      parse += len;
#ifdef MSG_LENGTH
      cout << "pushing msg durtn=" << con_msg.durtn << endl;
      cout << "           sender=" << con_msg.sender << endl;
      cout << "           recver=" << con_msg.recver << endl;
      cout << "data size =" << len << endl;
      if (len < MSG_LENGTH)
	cout << ">" << con_msg.message << "<" << endl;
#endif
      addMsg(lst, con_msg);
    }
  }
}

const SoccerMessage &MsgStore::getNextMessage() throw (Exception)
{
  if (curTime < -1)
    return SoccerMessage::emptyMessage;
#ifdef DEBUG
  cout << "get next msg for time " << curTime << endl;
#endif
  if (curMsg == NULL) {
#ifdef DEBUG
    cout << "using fall back" << endl;
#endif
    // had no more messages
    // now check if new ones are available
    if (*curMsgLoc == NULL)
      // no, got none
      return SoccerMessage::emptyMessage;
    // ok, got one
    curMsg = *curMsgLoc;
  }

  MsgItem *old = curMsg;
  curMsg    = old->next;
  curMsgLoc = &old->next;
#ifdef DEBUG
  cout << "finished. next msg = " << curMsg << endl;
  cout << "returning msg " << old->msg.sender
       << " " << old->msg.recver
       << " " << old->msg.message.length() << endl;
  if (old->msg.message.length() > 3000)
    *(int *)0 = 0;
#endif
  return old->msg;
}

void MsgStore::addMsg(MsgStore::MsgList &lst, const SoccerMessage &msg)
{
  MsgItem *newItem = new MsgItem;
  newItem->next = NULL;
  newItem->msg  = msg;

#ifdef DEBUG
  cout << "added msg at " << newItem << " of length "
       << newItem->msg.message.length()
       << " to list start " << lst.head << " list tail "
       << lst.tail << endl;
#endif

  // check for list empty
  if (lst.head == NULL) {
    lst.tail = lst.head = newItem;
    return;
  }
  lst.tail->next = newItem;
  lst.tail = newItem;
}

void MsgStore::clearList(MsgStore::MsgList &list)
{
  MsgItem *it = list.head;
  while (it) {
    MsgItem *tmp = it->next;
#ifdef DEBUG
    cout << "deleted " << it << endl;
#endif
    delete it;
    it = tmp;
  }
  list.head = list.tail = NULL;
}

int MsgStore::timeAddr(int t)
{
  int tmp = t % Buffered;
  if (tmp >= 0)
    return tmp;
  return Buffered + tmp;
}

void MsgStore::show()
{
  static const char *actor[NumberSActors] = {
    "***",
    "L 1", "L 2", "L 3",
    "L 4", "L 5", "L 6",
    "L 7", "L 8", "L 9",
    "L10", "L11", "LCC",

    "R 1", "R 2", "R 3",
    "R 4", "R 5", "R 6",
    "R 7", "R 8", "R 9",
    "R10", "R11", "RCc",

    "OlC", "SSv", "SMn", "FMn"
  };

  cout << "the logfile contains:" << endl;
  for (int t = -1;; t++) {
    if (!fetchTime(t))
      return;
    SoccerMessage s_msg;
    while ((s_msg = getNextMessage()).isValid()) {
      // avoid that bin messages
      if ((s_msg.recver != FunsselMonitor) ||
	  (s_msg.sender != SoccerServer)) {
	cout << setw(4) << t;
	if (s_msg.time != t)
	  cout << "[" << setw(4) << s_msg.time << "]";
	if (s_msg.sender != SoccerServer)
	  cout << " from " << setw(2) << actor[s_msg.sender];
	if (s_msg.recver != SoccerServer)
	  cout << " to " << setw(2) << actor[s_msg.recver];
	cout << ": " << string(s_msg) << endl;
      }
    }
  }
}
