//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//   $Id: grade.h,v 1.11 2001/09/11 10:01:37 goetz Exp $	
//   $Manager: cm$
//
//   grade.h
//
////////////////////////////////////////////////////////////////////////

#ifndef _GRADE_H_
#define _GRADE_H_

#include <iostream.h>
#include "audiomsg.h"

// The class "Grade" stores and manipulates valuations evaluations modelled
// after (German) school grades. It is e.g. used to evaluate agent actions.
// "Grade 0" means absoluteness, "Grade 0" impossibility. The different
// grades are not displayed as numbers but with words:
//              0 <-> ABSOLUTE
//              1 <-> PERFECT
//              2 <-> GOOD
//              3 <-> AVERAGE
//              4 <-> POOR
//              5 <-> BAD
//              6 <-> IMPOSSIBLE
// CAUTION: IMPOSSIBLE is no bad grade but means impossibility.
// The integer value (on the left side) is private, the different grades have
// their own methods. (This should prevent "adding" of grades or other
// weird things.)
// Values not in {0,...,6} leave the grade undefined. Comparison with
// undefined grades will always return false.
// CAUTION: There are no OPERATORS for comparison. (Not good in a system where
// higher numbers mean worse grades.)

class Grade
{
public:
  Grade() : value( -1 ) {}

  enum GradeConstant {
    IMPOSSIBLE = 6,
    BAD        = 5,
    POOR       = 4,
    AVERAGE    = 3,
    GOOD       = 2,
    PERFECT    = 1,
    ABSOLUTE   = 0
  };

  Grade(GradeConstant _v) : value( (int)_v ) {}

  // methods for manipulation

  void set_ABSOLUTE()
  { value = 0; }

  void set_PERFECT()
  { value = 1; }

  void set_GOOD()
  { value = 2; }

  void set_AVERAGE()
  { value = 3; }

  void set_POOR()
  { value = 4; }

  void set_BAD()
  { value = 5; }

  void set_IMPOSSIBLE()
  { value = 6; }

  // upgrade (CAUTION: PERFECT can not be upgraded to ABSOLUTE, IMPOSSIBLE
  // can not be upgraded to BAD)
  bool upgrade()
  { if( value > 1 && value <= 5 ) { value--; return true; } return false; }

  // downgrade (CAUTION: ABSOLUTE can not be downgraded to PERFECT, BAD
  // can not be downgraded to IMPOSSIBLE)
  bool downgrade()
  { if( value >= 1 && value < 5 ) { value++; return true; } return false; }
    
  // methods for query and comparison

  bool is_defined() const
  { return value >= 0 && value <= 6 ? true : false; }

  bool is_ABSOLUTE() const
  { return value == 0 ? true : false; }

  bool is_PERFECT() const
  { return value == 1 ? true : false; }

  bool is_GOOD() const
  { return value == 2 ? true : false; }

  bool is_AVERAGE() const
  { return value == 3 ? true : false; }

  bool is_POOR() const
  { return value == 4 ? true : false; }

  bool is_BAD() const
  { return value == 5 ? true : false; }

  bool is_IMPOSSIBLE() const
  { return value == 6 ? true : false; }

  bool is_like( const Grade& g ) const
  { return is_defined() && g.is_defined() && value == g.value ? true : false; }

  bool is_better_than( const Grade& g ) const
  { return is_defined() && g.is_defined() && value < g.value ? true : false; }

  bool is_worse_than( const Grade& g ) const
  { return is_defined() && g.is_defined() && value > g.value ? true : false; }

private:
  
  // comparison operator made private (so that you don't use it)
  bool operator == ( const Grade& g ) const
  { return is_defined() && g.is_defined() && value == g.value ? true : false; }

  int value;
};


// output

inline
class ostream & operator << ( class ostream &os, const Grade &g )
{
  return os << ( g.is_ABSOLUTE() ? "ABSOLUTE" :
		 g.is_PERFECT() ? "PERFECT" :
		 g.is_GOOD() ? "GOOD" :
		 g.is_AVERAGE() ? "AVERAGE" :
		 g.is_POOR() ? "POOR" :
		 g.is_BAD() ? "BAD" :
		 g.is_IMPOSSIBLE() ? "IMPOSSIBLE" :
		 "UNDEFINED" );
}

// operators for writing and reading of grades to AudioMsg's

inline
class AudioMsg& operator << ( AudioMsg& aumsg, const Grade &g )
{
  return aumsg << NodaChar( g.is_ABSOLUTE() ? char(7) : 
		 g.is_PERFECT() ? char(6) :
		 g.is_GOOD() ? char(5) :
		 g.is_AVERAGE() ? char(4) :
		 g.is_POOR() ? char(3) :
		 g.is_BAD() ? char(2) :
		 g.is_IMPOSSIBLE() ? char(1) :
		 char(0) );
}

inline
class AudioMsg& operator >> ( AudioMsg& aumsg, Grade &g )
{
  NodaChar n;
  aumsg >> n;
  switch (char(n)) {
  case 7: 
    g.set_ABSOLUTE(); break;
  case 6:
    g.set_PERFECT(); break;
  case 5:
    g.set_GOOD(); break;
  case 4:
    g.set_AVERAGE(); break;
  case 3:
    g.set_POOR(); break;
  case 2:
    g.set_BAD(); break;
  default:
    g.set_IMPOSSIBLE();
  }
  return aumsg;
}

#endif
