//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// libMRB Comm.C 
//
// $Id: Comm.C,v 1.21 2002/06/07 12:14:27 cs Exp $
// $manager: cs$
//
//////////////////////////////////////////////////////////////////////

#include <errno.h>
#include "Comm.h"
#include "FunsselStrings.h"

#ifndef NOCORE
#include <sys/resource.h>
#endif

#define CERR(msg) cerr << "CommError: " << msg << endl
#define INITTRACE(cmd)
#define COMMTRACE(cmd)
#define SAYTRACE(cmd)

Comm * Comm::theComm = NULL;

Comm::Comm(const char* _team, client_type ctype,
	   const char * host, int port, int reconnectNo,
	   bool FUNSSELmode, int clangMinVer, int clangMaxVer,
	   playing_side coachSide ) 
  : sock(NULL),
    parser(NULL),
    store(NULL),
    async(NULL),
    audiomgr(NULL),
    FUNSSELmode(FUNSSELmode),
    serverOK(true),
    reconnectNumber(reconnectNo),
    ctype(ctype),
    coachSide(coachSide)
{
  if (Comm::theComm == NULL) { // only if no other Comm exists yet
    Comm::theComm = this; // store static reference

    try {
      sock = new UDPSocket(host, port, MAX_SERVER_MSG_SIZE);
    }
    catch (UDPSocketError e) {
      CERR(e.msg());
      serverOK = false;
    }

    async    = new AsyncCommInfo(_team, ctype);
    parser   = new Parser(_team, ctype, clangMinVer, clangMaxVer);
    store    = new SensorDataStore();
    audiomgr = new AudioMgr();
    
  } else
    CERR("Comm: cannot construct more than one Comm "
	 << "in any single process !");

#ifndef NOCORE
  struct rlimit rlim;
  getrlimit(RLIMIT_CORE, &rlim);
  rlim.rlim_cur = rlim.rlim_max;
  cerr << "core dump enabled." << endl;  
  if (setrlimit(RLIMIT_CORE, &rlim) == -1)
    cerr << errno << endl;
#endif
}


Comm::~Comm() {
  if (serverOK) {
    char * msg = "(bye)";
    sock->send(msg);
  }

  Comm::theComm = NULL;
  
  delete sock;
  delete async;
  delete parser;
  delete store;
  delete audiomgr;
}

void 
Comm::drainSocket( void ) {

  const char * msg;

  if (*async->isActionNeeded()) {
    COMMTRACE(cout << "Comm: action needed returning from drainSocket 1" << endl);
    return;
  }

  try {
    if (!FUNSSELmode)
      // if we use the soccerserver it might crash (or finish), so don`t wait forever
      msg = sock->recv(SIMULATOR_STEP*COMM_MAX_DEAD_SERVER_CYCLES, async->isActionNeeded());
    else
      // if we use FUNSSEL we want to wait forever, as the server may be stopped
      msg = sock->recv(-1, async->isActionNeeded());
  }
  catch (UDPTimeOut &e) {
    serverOK = false;
    CERR(" (" << async->getTeam()
	 << " No. " << async->getNumber()
	 << ") " << e.msg() << " assuming server is dead");
    return;
  }
  catch (UDPAct &) {
    return;
  }
  catch (UDPSocketError &e) {
    if (e.number() != EINTR) {
      serverOK = false;
      CERR(" (" << async->getTeam()
	   << " No. " << async->getNumber()
	   << ") " << e.msg() << " assuming server is dead");
      return;
    }
    // here select has been interrupted by SIGALRM
    // so we probably want to act NOW
    return;
  }

  // read as long as socket has data
  while (true) {

    // set time of msg arrival
    async->setRecvTime();

    SensorData data;
    if (!parser->parseServerString(msg, data)) {
      CERR(getServerTime() 
	   << " Comm (" << async->getTeam() << " No." << async->getNumber()
	   << "): could not parse server string " << msg);
    } else {
      
      // use data to adjust time, play mode etc.
      bool firstSenseBody = async->adjust(data); 

      switch (data.type) {

      case SENSOR_TYPE_OK:
	COMMTRACE(cout << "Comm: skipping ok message" << endl);
	break;
      case SENSOR_TYPE_ERROR: 
	CERR(getServerTime()
	     << " Comm (" << async->getTeam() << " No." << async->getNumber()
	   << "): received error message from server: " << msg );
	break;
      case SENSOR_TYPE_WARNING: 
	CERR(getServerTime()
	     << " Comm (" << async->getTeam() << " No." << async->getNumber()
	   << "): received warning message from server: " << msg );
	break;
      case SENSOR_TYPE_STRANGE: 
	CERR(getServerTime()
	     << " Comm (" << async->getTeam() << " No." << async->getNumber()
	   << "): received strange message from server: " << msg );
	break;
      case SENSOR_TYPE_CLANG_VER:
	break;
      case SENSOR_TYPE_SENSE_BODY:
	if (async->isTimeRunning() && firstSenseBody) {
	  // first sense_body of a timestep received while clock is running:
	  // insert it before all data from same timestep 
	  // (needed by worldmodel)
	  COMMTRACE(cout << "Comm: inserting first sense_body data into store" 
		<< endl);
	  store->insert(data);
	} else {
	  // additional sense_body, or sense_body received while clock 
	  // is stopped: just store it according to receive time
	  COMMTRACE(cout << "Comm: storing sense_body data" << endl);
	  store->put(data);
	}
	break;
      case SENSOR_TYPE_EXACT:
	COMMTRACE(cout << "Comm: storing exact data" << endl);
	store->put(data);
	break;
      case SENSOR_TYPE_SEE:
	COMMTRACE(cout << "Comm: storing see data" << endl);
	store->put(data); // worldmodel can handle twisted neck
	break;
      case SENSOR_TYPE_PLAYER_TYPE:
	COMMTRACE(cout << "Comm: player_type" << endl);
	COMMTRACE(cout << "Comm: storing player_type data" << endl);
	store->put(data);
	break;
      case SENSOR_TYPE_PLAYER_CHANGED:
	COMMTRACE(cout << "Comm: player changed" << endl);
	COMMTRACE(cout << "Comm: storing player_changed data" << endl);
	store->put(data);
	break;
      case SENSOR_TYPE_HEAR:
	
	switch (data.sender) {
	case SENDER_SELF: 
	  COMMTRACE(cout << "Comm: skipping own message" << endl);
	  break;
	case SENDER_REFEREE:
	  COMMTRACE(cout << "Comm: skipping referee play mode" << endl);
	  // information contained was already used in async->adjust() above, 
	  // so also forget about it
	  break;
	case SENDER_DIR:    // message from player heard by player
	  COMMTRACE(cout << "Comm: storing heard message" << endl);
	  SAYTRACE(cout << "Comm: decoding via AudioMgr " 
		   << data.message << endl);
	  audiomgr->storeMsg(AudioMsg(data.message));
	  break;
	case SENDER_PLAYER: // message from player heard by any coach
	  COMMTRACE(cout << "Comm: storing player message" << endl);
	  SAYTRACE(cout << "Comm: decoding via AudioMgr " 
		   << data.message << endl);
	  audiomgr->storeMsg(AudioMsg(data.message));
	  break;
	case SENDER_PLAYER_US: // message from our player heard by any player
	  COMMTRACE(cout << "Comm: storing player message" << endl);
	  SAYTRACE(cout << "Comm: decoding via AudioMgr " 
		   << data.message << endl);

	  // put message into store (soccerserver 8+ communication)
	  store->put(data);

	  // old code: put message into audiomgr
	  // audiomgr->storeMsg(AudioMsg(data.message));
	  break;
	case SENDER_PLAYER_THEM: // message from other player heard by any player
	  COMMTRACE(cout << "Comm: skipping other player message" << endl);
	  break;
	case SENDER_COACH:  // message from train coach heard by player/olcoach
	  if (async->isTrainCoach()) {
	    COMMTRACE(cout << "Comm: skipping own message" << endl);
	  } else {
	    COMMTRACE(cout << "Comm: storing coach message relayed "
		  << "through the referee" << endl);
	    SAYTRACE(cout << "Comm: decoding via AudioMgr " 
		     << data.message << endl);
	    audiomgr->storeMsg(AudioMsg(data.message));
	  }
	  break;
	case SENDER_ONLINECOACH:// msg from onlinecoach heard by player/coach
	  COMMTRACE (cout << "Comm: recvd message from olcoach" << endl);
	  if (async->isOnlineCoach()) {
	    COMMTRACE(cout << "Comm: skipping own message" << endl);
	  } else {
	    COMMTRACE(cout << "Comm: storing online coach message" << endl);
	    switch (data.coachMsgType) {
	    case CMT_FREEFORM: {
	      SAYTRACE(cout << "Comm: decoding via AudioMgr " 
		       << data.message << endl);
	      audiomgr->storeMsg(AudioMsg(data.message));
	      break;
	    }
	    case CMT_META:
	    case CMT_INFO:
	    case CMT_ADVICE:
	    case CMT_DEFINE:
	    default:;
	    } // switch on CMT
	  } // !isOnlineCoach
	  break;
	} // switch on sender
	break;
      } //switch on sensor_type
    } //parse ok

    // this is to avoid reading msgs till the next cycle
    if (*async->isActionNeeded()) {
      COMMTRACE(cout << "Comm: action needed returning from drainSocket 2" << endl);
      return;
    }

    try {
      msg = sock->recv(0, async->isActionNeeded());
    }
    catch (UDPTimeOut &) {
      // no more data in socket
      COMMTRACE(cout << "Comm: no more data in socket" << endl);
      return;
    }
    catch (UDPAct &) {
      return;
    }
    catch (UDPSocketError &e) {
      if (e.number() != EINTR) {
	serverOK = false;
	CERR(" (" << async->getTeam()
	     << " No. " << async->getNumber()
	     << ") " << e.msg() << " assuming server is dead");
	return;
      }
      // here select has been interrupted by SIGALRM
      // so we probably want to act NOW
      return;
    }
  } // while true
}

void
Comm::generateCheckinMessage( char * outmsg )
{
  if (reconnectNumber <= 0) { // player first init or any coach

    switch (ctype) {

    case CLIENT_GOALIE:
      sprintf(outmsg, "(init %s (version 8.04) (goalie))", 
	      async->getTeam());
      break;

    case CLIENT_FIELDPLAYER:
      sprintf(outmsg, "(init %s (version 8.04))", async->getTeam());
      break;

    case CLIENT_ONLINECOACH:
      sprintf(outmsg, "(init %s (version 8.04))", async->getTeam());
      break;

    case CLIENT_TRAINCOACH:
      sprintf(outmsg, "(init (version 8.04))");
      break;

    default: 
      break;
    }
  } else {
    switch (ctype) {

    case CLIENT_GOALIE:
    case CLIENT_FIELDPLAYER:
      sprintf(outmsg, "(reconnect %s %d)", async->getTeam(), reconnectNumber);
      break;

    case CLIENT_ONLINECOACH:
    case CLIENT_TRAINCOACH:
      ensure(reconnectNumber == 0); // coach cannot reconnect
      break;

    default:
      break;
    }
  } 
}

comm_error
Comm::parseCheckinResponse( const char * inmsg )
{
  char   c;
  char   playmode[MAX_SERVER_MSG_SIZE];

  int          time;
  bool         secondHalf;
  int          number = -1;
  play_mode    mode   = PLAY_MODE_UNKNOWN;
  playing_side side   = SIDE_UNKNOWN;
  
  if (reconnectNumber <= 0) { // parse the init response
    time = 0;
    secondHalf = false;
    switch (ctype) {

    case CLIENT_GOALIE:
    case CLIENT_FIELDPLAYER:
      if (sscanf(inmsg, "(init %c %d %[^)])", &c, &number, playmode) != 3)
	return COMM_ERROR_PARSE;

      if (c == 'l')
	side = SIDE_LEFT;
      else if (c == 'r') 
	side = SIDE_RIGHT;
      else 
	return COMM_ERROR_PARSE;

      mode = Parser::parsePlaymode(playmode, side);
      break;

    case CLIENT_ONLINECOACH:
      if (sscanf(inmsg, "(init %c ok)" , &c) != 1) 
	return COMM_ERROR_PARSE;
      else {
	number = 0;
	if (c == 'l')
	  side = SIDE_LEFT;
	else if (c == 'r')
	  side = SIDE_RIGHT;
	else
	  return COMM_ERROR_PARSE;

	mode = PLAY_MODE_BEFORE_KICK_OFF;
	sendEye(true);
      }
      break;

    case CLIENT_TRAINCOACH:
      if (strcmp(inmsg,"(init ok)"))
	return COMM_ERROR_PARSE;
      else {
	number = 0;
	side = coachSide;
	if (coachSide == SIDE_LEFT)
	  mode = PLAY_MODE_KICK_OFF_US;
	else 
	  mode = PLAY_MODE_KICK_OFF_THEM;
	sendEar(true);
	sendEye(true);
      }
      break;

    default:
      break;
    }
  } else {  // parse the reconnect response
    time = 3000;
    secondHalf = true;
    switch (ctype) {

    case CLIENT_GOALIE:
    case CLIENT_FIELDPLAYER:
      if (sscanf(inmsg, "(reconnect %c %[^)])", &c, playmode) != 2)
	return COMM_ERROR_PARSE;

      number = reconnectNumber;
      if (c == 'l')
	side = SIDE_LEFT;
      else if (c == 'r')
	side = SIDE_RIGHT;
      else
	return COMM_ERROR_PARSE;

      mode = Parser::parsePlaymode(playmode, side);
      break;

    case CLIENT_ONLINECOACH:
    case CLIENT_TRAINCOACH:
      ensure(reconnectNumber == 0); // coach cannot reconnect
      break;
      
    default:
      break;
    }
  }
  
  async->setStates(number, time, mode, side, secondHalf);
  parser->setSide(side);
  
  return COMM_OK;
}

comm_error 
Comm::checkinAtServer( void )
{
  INITTRACE( cout << "Comm:     ";
	     switch(ctype) {
	     case CLIENT_GOALIE: cout << "goalie "; break;
	     case CLIENT_FIELDPLAYER: cout << "fieldplayer "; break;
	     case CLIENT_ONLINECOACH: cout << "onlinecoach "; break;
	     case CLIENT_TRAINCOACH: cout << "trainingcoach "; break;
	     default: break;
	     }
	     if (reconnectNumber > 0) 
	        cout << "reconnecting" << endl;
	     else 
	     cout << "checking into server (CL-Version: "
	     << parser->getCLangMinVer() << " "
	     << parser->getCLangMaxVer() << ")" << endl;
	     );
  
  char outmsg[MAX_SERVER_MSG_SIZE];
  const char * inmsg;

  generateCheckinMessage(outmsg);
  
  // send message to server
  INITTRACE (cout << "Comm:     sending checkin message " 
	     << outmsg << " to server" << endl);

  // send init
  try {
    sock->send(outmsg);
  }
  catch (UDPSocketError e) {
    CERR(e.msg());
    return COMM_ERROR_SERVER_INIT;
  }
  
  INITTRACE( cout << "Comm:     waiting for server response" << endl);
  
  try {
    inmsg = sock->recv(SIMULATOR_STEP*COMM_MAX_DEAD_SERVER_CYCLES);
  }
  catch (UDPTimeOut e) {
    CERR(e.msg());
    return COMM_ERROR_NO_DATA;
  }
  catch (UDPSocketError e) {
    CERR(e.msg());
    return COMM_ERROR_SERVER_INIT;
  }

  INITTRACE (cout << "Comm:     received server response " 
	     << inmsg << endl);
  
  comm_error ce;
  ce = parseCheckinResponse(inmsg);
  if (ce != COMM_OK) {
    CERR("Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): cannot handle server checkin response " << inmsg );
    return ce;
  }

  // send cl-version
  if (ctype == CLIENT_GOALIE || ctype == CLIENT_FIELDPLAYER) {
    sprintf(outmsg, "(clang (ver %d %d))",
	    parser->getCLangMinVer(),
	    parser->getCLangMaxVer());
    
    INITTRACE (cout << "Comm:     sending clang message " 
	       << outmsg << " to server" << endl);

    try {
      sock->send(outmsg);
    }
    catch (UDPSocketError e) {
      CERR(e.msg());
      return COMM_ERROR_SERVER_INIT;
    }

    /* do not wait for (ok clang...) as this is sent after player_type(s)
       and is therefore handled by the parser (droped) */
  }
  
  INITTRACE( cout << "Comm:     waiting for server_param" << endl);

  try {
    inmsg = sock->recv(SIMULATOR_STEP*COMM_MAX_DEAD_SERVER_CYCLES);
  }
  catch (UDPTimeOut e) {
    CERR(e.msg());
    return COMM_ERROR_NO_DATA;
  }
  catch (UDPSocketError e) {
    CERR(e.msg());
    return COMM_ERROR_SERVER_INIT;
  }
  
  if (!parser->parseServerParam(inmsg)) {
    CERR("Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): cannot parse server_param " << inmsg );
    return COMM_ERROR_SERVER_INIT;
  }

  INITTRACE( cout << "Comm:     waiting for player_param" << endl);

  try {
    inmsg = sock->recv(SIMULATOR_STEP*COMM_MAX_DEAD_SERVER_CYCLES);
  }
  catch (UDPTimeOut e) {
    CERR(e.msg());
    return COMM_ERROR_NO_DATA;
  }
  catch (UDPSocketError e) {
    CERR(e.msg());
    return COMM_ERROR_SERVER_INIT;
  }

  if (!parser->parsePlayerParam(inmsg)) {
    CERR("Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): cannot parse player_param " << inmsg );
    return COMM_ERROR_SERVER_INIT;
  }

  INITTRACE( cout << "Comm:     completed checkin at sserver" << endl);

#if 1
  if (ctype != CLIENT_ONLINECOACH && ctype != CLIENT_TRAINCOACH) {
    sprintf(outmsg,"(ear (off opp))");

    try {
      sock->send(outmsg);
    }
    catch (UDPSocketError e) {
      CERR(e.msg());
      return COMM_ERROR_SERVER_INIT;
    }
  }
#endif

  return COMM_OK;
}

comm_error 
Comm::connect( void )
{
  
  INITTRACE(cout << "Comm: connecting to server" << endl);

  comm_error e;
  
  e = checkinAtServer(); 
  if (e) 
    return e;

  // set parameters known after receiving first info

  audiomgr->setStates(async->getNumber(), async->getSide() == SIDE_LEFT );
  
  serverOK = true;
  
  return COMM_OK;
}

void
Comm::coachSay( void /*???*/) {
}

void
Comm::sendMCmd( const MCommand &cmd )
{
  if (!Comm::theComm->FUNSSELmode) {
    CERR("Not in FUNSSELmode dropping cmd: \"" << cmd << "\"");
    return;
  }
  char msg[strlen(FILTER_STRING)+cmd.getLength()+10];
  sprintf(msg, FILTER_STRING " %d %s)", cmd.duration, ((string&)cmd).c_str());
  try {
    Comm::theComm->sock->send(msg);
  }
  catch (UDPSocketError e) {
    CERR(Comm::theComm->getServerTime() 
	 << " Comm (" << Comm::theComm->async->getTeam() << " No." 
	 << Comm::theComm->async->getNumber()
	 << "): cannot send mcmd \"" << cmd << "\" to server." );
    CERR(e.msg());
  }
}

void
Comm::say(AudioMsg &msg) {
  //[C]
  char charmsg[MAX_SAY_MSG_SIZE];

  if (msg.isCoachMsg()) {
    sprintf(charmsg, "(freeform \"%s\")", audiomgr->SpeakTo(msg));
//      strcpy(charmsg, "(freeform \""); 
//      strcat(charmsg, audiomgr->SpeakTo(msg));
//      strcat(charmsg, "\")");
  } else
    strcpy(charmsg, audiomgr->SpeakTo(msg));

  SAYTRACE(cout << "Comm: encoding via AudioMgr " << msg << endl);
  sendSay(charmsg);
}

void
Comm::say_10( char * msg )
{
  sendSay(msg);
}

void 
Comm::noAction( void )
{
  COMMTRACE(cout << "Comm: agent doing no action" << endl);
  async->setSentAction(NO_COMMAND, async->getServerTime(), 0, 0);
}

void 
Comm::sendKick( double power, double dir_rad )
{ 

  double dir_deg = RAD_TO_DEG( dir_rad );

  if (power < MINPOWER - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), normalizing to " << MINPOWER );
    power = MINPOWER;
  } else if (power > MAXPOWER + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), normalizing to " << MAXPOWER );
    power = MAXPOWER;
  } else if (dir_deg < MINMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), normalizing to " << MINMOMENT );
    dir_deg = MINMOMENT;
  } else if (dir_deg > MAXMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), normalizing to " << MAXMOMENT );
    dir_deg = MAXMOMENT;
  } else if (! ( (power >= MINPOWER - EPS) && 
		 (power <= MAXPOWER + EPS) && 
		 (dir_deg >= MINMOMENT - EPS) && 
		 (dir_deg <= MAXMOMENT + EPS) ) ) {
    // NaN somewhere ?
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), dropping command completely" );
    return;
  }
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(kick %f %f)", power, dir_deg);

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
    async->setSentAction(KICK_COMMAND, async->getServerTime(), 
			power, DEG_TO_RAD( dir_deg ));
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendDash( double power )
{ 

  if (power < MINPOWER - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (dash " << power 
	 << "), normalizing to " << MINPOWER );
    power = MINPOWER;
  } else if (power > MAXPOWER + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (dash " << power 
	 << "), normalizing to " << MAXPOWER );
    power = MAXPOWER;
  } else if (! ( (power >= MINPOWER - EPS) && 
		 (power <= MAXPOWER + EPS) ) ) {
    // NaN ?
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (dash " << power 
	 << "), dropping command completely" );
    return;
  }
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(dash %f)", power);

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
    async->setSentAction(DASH_COMMAND, async->getServerTime(), power, 0);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendTurn( double moment_rad )
{ 

  double moment_deg = RAD_TO_DEG( moment_rad );

  if (moment_deg < MINMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (turn " << moment_deg 
	 << "), normalizing to " << MINMOMENT );
    moment_deg = MINMOMENT;
  } else if (moment_deg > MAXMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (turn " << moment_deg 
	 << "), normalizing to " << MINMOMENT );
    moment_deg = MAXMOMENT;
  } else if ( ! ( (moment_deg >= MINMOMENT - EPS) && 
		  (moment_deg <= MAXMOMENT + EPS) ) ) {
    // NaN ?
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (turn " << moment_deg 
	 << "), dropping command completely" );
    return;
  }
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(turn %f)", moment_deg);

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
    async->setSentAction(TURN_COMMAND, async->getServerTime(), 
			moment_deg * M_PI / 180.0, 0);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendCatch( double dir_rad )
{ 

  double dir_deg = RAD_TO_DEG( dir_rad );

  if (dir_deg < MINMOMENT - EPS) {
    CERR(getServerTime()
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (catch " << dir_deg 
	 << "), normalizing to " << MINMOMENT );
    dir_deg = MINMOMENT;
  } else if (dir_deg > MAXMOMENT + EPS) {
    CERR(getServerTime()
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (catch " << dir_deg 
	 << "), normalizing to " << MAXMOMENT );
    dir_deg = MAXMOMENT;
  } else if (! ( (dir_deg >= MINMOMENT - EPS) && 
		 (dir_deg <= MAXMOMENT + EPS) ) ) {
    // NaN ?
    CERR(getServerTime()
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (catch " << dir_deg 
	 << "), dropping command completely" );
      return;
  }
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(catch %f)", dir_deg);

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
    async->setSentAction(CATCH_COMMAND, async->getServerTime(), 
			 DEG_TO_RAD( dir_deg ), 0);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendMove( double x, double y )
{ 

  if ( (!async->isGoalie() && async->isTimeRunning() ) ||
       (async->isGoalie() && async->isTimeRunning() && 
	(async->getPlayMode() != PLAY_MODE_FREE_KICK_US) ) ) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << ") game is running, agent not allowed to (move)" );
    return;
  } else {

    // no range checking when moving: there are no limits !
    
    char msg [MAX_SERVER_MSG_SIZE];
    sprintf(msg, "(move %f %f)", x, y);

    try {
      sock->send(msg);
      COMMTRACE(cout << "Comm: agent sent " << msg << endl);
      async->setSentAction(MOVE_COMMAND, async->getServerTime(), x, y);
    }
    catch (UDPSocketError e) {
      CERR(getServerTime() 
	   << " Comm (" << async->getTeam() << " No." << async->getNumber()
	   << "): agent cannot send " << msg << " to server" );
      CERR(e.msg());
    }
  } 
}

void 
Comm::sendTurnNeck( double angle_rad )
{ 

  double angle_deg = RAD_TO_DEG( angle_rad );

  if (angle_deg < MINNECKMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (turn_neck " << angle_deg 
	 << "), normalizing to " << MINNECKMOMENT );
    angle_deg = MINNECKMOMENT;
  } else if (angle_deg > MAXNECKMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (turn_neck " << angle_deg 
	 << "), normalizing to " << MAXNECKMOMENT );
    angle_deg = MAXNECKMOMENT;
  } else if (! ( (angle_deg >= MINNECKMOMENT - EPS) && 
		 (angle_deg <= MAXNECKMOMENT + EPS) ) ) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (turn_neck " << angle_deg 
	 << "), dropping command completely" );
    return;
  }
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(turn_neck %f)", angle_deg);

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
    async->incTurnNeckCount();
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendChangePlayer( int player, int type )
{
  char msg[50];
  sprintf(msg, "(change_player_type %d %d)", player, type);

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void
Comm::sendChangeView( view_width width, view_quality quality )
{
  char w[16];
  switch (width) {
  case VIEW_WIDTH_NARROW: strcpy(w, "narrow"); break;
  case VIEW_WIDTH_NORMAL: strcpy(w, "normal"); break;
  case VIEW_WIDTH_WIDE: strcpy(w, "wide"); break;
  case VIEW_WIDTH_NONE: strcpy(w, "normal"); break; 
  }

  char q[8];
  switch (quality) {
  case VIEW_QUALITY_LOW: strcpy(q, "low"); break;
  case VIEW_QUALITY_HIGH: strcpy(q, "high"); break;
  }

  char msg[MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(change_view %s %s)", w, q);


  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
    async->incChangeViewCount();
    async->setViewMode(width, quality);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendPointTo( double dist, double dir_rad )
{ 
  double dir_deg = RAD_TO_DEG( dir_rad );

  char msg[MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(pointto %f %f)", dist, dir_deg);

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
    async->incPointToCount();
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendPointToOff( void )
{ 
  char * msg = "(pointto off)";

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
    async->incPointToCount();
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void
Comm::sendAttentionTo( int pNum, bool ourTeam ) {
  char msg[MAX_SERVER_MSG_SIZE];

  if (pNum < 0 || pNum > 11 ||
      (ourTeam && pNum == async->getNumber())) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (attentionto " << pNum << " " 
	 << (ourTeam?"our":"opp") << "), dropping command completely" );
    return;
  }

  sprintf(msg, "(attentionto %d %s)", pNum, ourTeam?"our":"opp");

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
    async->incAttentionToCount();
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendAttentionToOff( void )
{ 
  char * msg = "(attentionto off)";

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
    async->incAttentionToCount();
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void
Comm::sendTackle( double power ) {
  char msg[MAX_SERVER_MSG_SIZE];

//    if (power < MINPOWER - EPS) {
//      CERR(getServerTime() 
//  	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
//  	 << "): agent tries to send (tackle " << power << " " 
//  	 << "), normalizing to " << MINPOWER );
//      power = MINPOWER;
//    } else if (power > MAXPOWER + EPS) {
//      CERR(getServerTime() 
//  	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
//  	 << "): agent tries to send (tackle " << power << " " 
//  	 << "), normalizing to " << MAXPOWER );
//      power = MAXPOWER;
//    }
  
  sprintf(msg, "(tackle %f)", power);

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
    async->setSentAction(TACKLE_COMMAND, async->getServerTime(), power, 0);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendSenseBody( void )
{
  char * msg = "(sense_body)";
  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendSay( const char * message )
{ 
  unsigned int maxSize = MAX_SAY_MSG_SIZE+1;
  if (async->isCoach()) 
    // + strlen("(freeform")+strlen(")") !!
    maxSize = SAY_COACH_MSG_SIZE + 12; 

  if (strlen(message) > maxSize) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): client tries to send too long say message (len= "
	 << strlen(message) << ")(msg= " << message 
	 << "), dropping it completely" );
    return;
  }
  
  //  cerr << "sendSay: " << message << endl;


  char msg[maxSize+9];
  if (async->isCoach())
    sprintf(msg, "(say %s)", message);
  else
    sprintf(msg, "(say \"%s\")", message);

  //  cerr << "sendSay2: " << msg << endl;

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
    async->incSayCount();
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendDone( void )
{
  char * msg = "(done)";
  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendChangeMode( play_mode playMode )
{ 
  char * newMode;
  switch (playMode) {
  case PLAY_MODE_BEFORE_KICK_OFF: newMode = "before_kick_off"; break;
  case PLAY_MODE_CORNER_KICK_US: 
    if (async->getSide()==SIDE_LEFT) newMode = "corner_kick_l";
    else newMode = "corner_kick_r";
    break;
  case PLAY_MODE_CORNER_KICK_THEM: 
    if (async->getSide()==SIDE_LEFT) newMode = "corner_kick_r";
    else newMode = "corner_kick_l";
    break;
  case PLAY_MODE_FREE_KICK_US: 
    if (async->getSide()==SIDE_LEFT) newMode = "free_kick_l";
    else newMode = "free_kick_r";
    break;
  case PLAY_MODE_FREE_KICK_THEM: 
    if (async->getSide()==SIDE_LEFT) newMode = "free_kick_r";
    else newMode = "free_kick_l";
    break;
  case PLAY_MODE_GOAL_KICK_US: 
    if (async->getSide()==SIDE_LEFT) newMode = "goal_kick_l";
    else newMode = "goal_kick_r";
    break;
  case PLAY_MODE_GOAL_KICK_THEM: 
    if (async->getSide()==SIDE_LEFT) newMode = "goal_kick_r";
    else newMode = "goal_kick_l";
    break;
  case PLAY_MODE_KICK_IN_US: 
    if (async->getSide()==SIDE_LEFT) newMode = "kick_in_l";
    else newMode = "kick_in_r";
    break;
  case PLAY_MODE_KICK_IN_THEM: 
    if (async->getSide()==SIDE_LEFT) newMode = "kick_in_r";
    else newMode = "kick_in_l";
    break;
  case PLAY_MODE_KICK_OFF_US: 
    if (async->getSide()==SIDE_LEFT) newMode = "kick_off_l";
    else newMode = "kick_off_r";
    break;
  case PLAY_MODE_KICK_OFF_THEM: 
    if (async->getSide()==SIDE_LEFT) newMode = "kick_off_r";
    else newMode = "kick_off_l";
    break;
  case PLAY_MODE_GOAL_US: 
    if (async->getSide()==SIDE_LEFT) newMode = "goal_l";
    else newMode = "goal_r";
    break;
  case PLAY_MODE_GOAL_THEM: 
    if (async->getSide()==SIDE_LEFT) newMode = "goal_r";
    else newMode = "goal_l";
    break;
  case PLAY_MODE_OFFSIDE_US: 
    if (async->getSide()==SIDE_LEFT) newMode = "offside_l";
    else newMode = "offside_r";
    break;
  case PLAY_MODE_OFFSIDE_THEM: 
    if (async->getSide()==SIDE_LEFT) newMode = "offside_r";
    else newMode = "offside_l";
    break;
  case PLAY_MODE_PLAY_ON : newMode = "play_on"; break;
  case PLAY_MODE_TIME_OVER : newMode = "time_over"; break;
  default: 
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): trying to set unknown play mode: No." 
	 << (int)playMode << ", droping command" ); 
    return;
  }
  
  char msg[MAX_SERVER_MSG_SIZE];
  sprintf (msg, "(change_mode %s)", newMode);

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendMoveObject( const char * object, double x, double y )
{ 
  char msg[MAX_SERVER_MSG_SIZE];
  sprintf (msg, "(move %s %f %f)", object, x, y);

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendMoveObject( const char * object, 
		      double x, double y, 
		      double dir_rad )
{ 
  double dir_deg = RAD_TO_DEG( dir_rad );

  if (dir_deg < MINMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), normalizing to " << MINMOMENT );
    dir_deg = MINMOMENT;
  } else if (dir_deg > MAXMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), normalizing to " << MAXMOMENT );
    dir_deg = MAXMOMENT;
  } else if (! ( (dir_deg >= MINMOMENT - EPS) && 
		 (dir_deg <= MAXMOMENT + EPS) ) ) {
    // NaN ?
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), dropping command completely" );
    return;
  }
  
  char msg[MAX_SERVER_MSG_SIZE];
  sprintf (msg, "(move %s %f %f %f)", object, x, y, dir_deg);

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendMoveObject( const char *object, double x, double y, 
		      double dir_rad, double xvel, double yvel )
{ 
  double dir_deg = RAD_TO_DEG( dir_rad );

  if (dir_deg < MINMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), normalizing to " << MINMOMENT );
    dir_deg = MINMOMENT;
  } else if (dir_deg > MAXMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), normalizing to " << MAXMOMENT );
    dir_deg = MAXMOMENT;
  } else if (! ( (dir_deg >= MINMOMENT - EPS) && 
		 (dir_deg <= MAXMOMENT + EPS) ) ) {
    // NaN ?
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), dropping command completely" );
    return;
  }
  
  char msg[MAX_SERVER_MSG_SIZE];
  sprintf (msg, "(move %s %f %f %f %f %f)", object, x, y, dir_deg, xvel, yvel);

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendEye( bool eyeOn )
{
  char * msg;
  if (eyeOn)
    msg = "(eye on)";
  else
    msg = "(eye off)";

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendEar( bool earOn )
{ 
  char * msg;
  if (earOn)
    msg = "(ear on)";
  else
    msg = "(ear off)";

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendLook( void )
{
  char * msg = "(look)";

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}  

void 
Comm::sendCheckBall( void )
{ 
  char * msg = "(check_ball)";

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}

void 
Comm::sendRecover( void )
{ 
  char * msg = "(recover)";

  try {
    sock->send(msg);
    COMMTRACE(cout << "Comm: agent sent " << msg << endl);
  }
  catch (UDPSocketError e) {
    CERR(getServerTime() 
	 << " Comm (" << async->getTeam() << " No." << async->getNumber()
	 << "): agent cannot send " << msg << " to server" );
    CERR(e.msg());
  }
}
