//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include <arpa/inet.h>
#include <netdb.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include <errno.h>
#include "UDPSocket.h"

UDPSocketError::UDPSocketError( const char * msg, int errNo ) 
  : errmsg(NULL), errNo(errNo)
{
  int len(strlen(msg)+1);
  errmsg = new char[len];
  strcpy(errmsg,msg);
};

UDPSocketError::UDPSocketError( const UDPSocketError &e )
  : errmsg(NULL),
    errNo(e.errNo)
{
  int len(strlen(e.errmsg)+1);
  errmsg = new char[len];
  strcpy(errmsg,e.errmsg);
}

UDPSocketError::~UDPSocketError( void )
{
  delete errmsg;
};

const char * 
UDPSocketError::msg( void )
{
  return errmsg;
};

int
UDPSocketError::number( void )
{
  return errNo;
};

UDPTimeOut::UDPTimeOut( void )
  : UDPSocketError("select timed out", -1) 
{};

UDPTimeOut::UDPTimeOut( const UDPTimeOut &e )
  : UDPSocketError(e) 
{};

UDPTimeOut::~UDPTimeOut( void )
{};

UDPAct::UDPAct( void )
  : UDPSocketError("need to act", -1) 
{};

UDPAct::UDPAct( const UDPAct &e )
  : UDPSocketError(e) 
{};

UDPAct::~UDPAct( void )
{};

//////////////////////////////////////////////////////////////////////

UDPSocket::UDPSocket( const char * host, int port, size_t maxMsgSize ) 
  throw (UDPSocketError)
  : sd(-1),
    recMsg(NULL),
    maxMsgSize(maxMsgSize+1)
{
  if (maxMsgSize <= 0)
    throw UDPSocketError("maxMsgSize <= 0", -1);
  
  recMsg = new char[maxMsgSize];

  if ((sd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    throw UDPSocketError("could not create socket", -1);

  // construct server in_addr
  bzero(&srv, sizeof(srv));
  srv.sin_family = AF_INET;

  struct hostent *hp;
  hp = gethostbyname(host);
  if (!hp)
    throw UDPSocketError("server host unknown", -1);

  srv.sin_addr = *((in_addr *)hp->h_addr);
  srv.sin_port = htons(port);

  // bind socket to a port
  struct sockaddr_in saddr;
  bzero(&saddr, sizeof(saddr));
  saddr.sin_family      = AF_INET;
  saddr.sin_addr.s_addr = htonl(INADDR_ANY);
  saddr.sin_port        = htons(0);
  if (bind(sd, (struct sockaddr *)&saddr, sizeof(saddr)) < 0)
    throw UDPSocketError("could not bind socket", errno);
}


UDPSocket::~UDPSocket( void ){
  delete recMsg;
  if (sd > -1)
    close(sd);
}

void
UDPSocket::send( const char * msg )
  throw (UDPSocketError)
{
  int len = strlen(msg)+1;
  int actual;
  // repeat sending if interrupted
  do {
    actual = sendto(sd, msg, len, 0, (struct sockaddr *) &srv, sizeof(srv));
  } while (actual < 0 && errno == EINTR);
  
  if (actual != len)
    throw UDPSocketError("could not send msg", -1);
}

const char *
UDPSocket::recv( int msec, const bool * act )
  throw (UDPSocketError, UDPTimeOut, UDPAct)
{
  recMsg[0] = 0;
  // create the file descriptor set to be watched by select
  fd_set  sset;
  struct timeval * timeout(NULL);
  struct timeval tv;

  FD_ZERO (&sset);
  FD_SET (sd, &sset);

  if (msec > -1) {
    tv.tv_sec  = msec/1000;
    tv.tv_usec = (msec%1000)*1000;
    timeout = &tv;
  }

  socklen_t fromlen;
  struct sockaddr_in from_addr;
  int n(-1);
  
  if (act && *act)
    throw UDPAct();

  switch (select(sd+1, &sset, NULL, NULL, timeout)) {
      
  case 1:
    fromlen = sizeof(from_addr) ;
    
    // repeat receiving if interrupted
    do {
      n = recvfrom(sd, recMsg, maxMsgSize-1, 0,
		   (struct sockaddr *)&from_addr, &fromlen);
    } while ((n == -1) && (errno == EINTR));
    
    if (n<0) 
      throw UDPSocketError("recvfrom returned without data", -1);
    
    recMsg[n] = 0;
    
    // sender of this message will be next adressee (when receiving init)
    srv.sin_port = from_addr.sin_port;
    break;
      
  case 0:
    throw UDPTimeOut();
    break;

  case -1:
    throw UDPSocketError("error while recving data", errno);
      break;
      
  default: 
    throw UDPSocketError("unknown return value from select", -2);
  } //switch
  
  return recMsg;
}
