//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// $Id: ballhandlingAct.C,v 1.90 2002/06/04 13:53:24 cm Exp $

// cwz-build-command: ssh chakotay "cd /nhomes/janeway/axel/RoboCup_2000; make"

/////////////////////////////////////////////////////////////////////////
//
// ACT
//
/////////////////////////////////////////////////////////////////////////


bool 
BallHandlingModule::act( AdvSkills* skills,
			 const Snapshot* snap,
			 TacticalSnapshot* tsnap ) {
  // main action method, called by master control to execute action
  // previously evaluated

  acted = true;

  switch (action) {
  case WAIT: {
    // if player is immobile after tackling, do not do anything
    if( snap->time < snap->lastTackleTime + TACKLE_CYCLES )
      skills->noAction();
    else
      {
	// turn perpetual to ball
	// to maximize reaction possibilities
	// choose that direction pointing into opponent goal
	// since we want to catch the ball later in that direction
	// if possible
	Angle dir = (ball.pos - my.pos).getAngle();
	dir += (dir > 0) ? -PI_2 : PI_2;
	AdvSkills::TurnCommand turnParams =
	  skills->turn(snap, dir);
	if ( (turnParams.state == AdvSkills::ROK) ||
	     (turnParams.state == AdvSkills::RANGLE) )
	  skills->exec(snap, turnParams);
	else
	  skills->noAction();
      }
    break;
  }
  case FIND_BALL:
    findBall();
    break;
  case GOALIE_INTERCEPT:
    updateTA(TA_CONTROL);
    playerIntercept();
    break;        
  case GOALIE_CATCH:
    updateTA(TA_CONTROL);
    goalieCatch();
    break;
  case GOALIE_DRIBBLE:
    updateTA(TA_CONTROL);
    fastdribble();
    break;  
  case GOALIE_PLACE:
    updateTA(TA_CONTROL);
    place();
    break;
  case PLAYER_INTERCEPT:
    updateTA(TA_CONTROL);
    playerIntercept();
    break;
  case PLAYER_INTERCEPT_ALT:
    // either going for a fallback position or
    // for a guessed ball position
    skills->exec(snap, turtleCmd);
    break;
  case PLAYER_FAST_DRIBBLE:
    updateTA(TA_CONTROL);
    fastdribble();
    break;
  case PLAYER_DRIBBLE:
    updateTA(TA_CONTROL);
    dribble();
    break;
  case PLAYER_PLACE:
    updateTA(TA_CONTROL);
    place();
    break;
  default:
    CERR ("act -> dont know what to do with action = " << (int)action);
    skills->noAction();
    return false;
  }
  return true; 
}


void
BallHandlingModule::updateTA(TacticalActionState ta)
{
  if ((ta != TA_REACT) && (ta != TA_CONTROL))
    return;

  // if any player of our team except for me is near the
  // ball, be quiet. I do not use the reachTime but real
  // distance since reachTime is instable and I don't want
  // to fill up our EmChannel with nonsense
  int clst = snap->ourTeamClosestTo(ball.pos);
  if ((clst != UNKNOWN) && (clst != my.number) &&
      (snap->ourTeam(clst).distTo(ball) < 2*snap->ourTeam(clst).vmax))
    return;

  // if the ta has not changed be quiet (for some cycles at least)
  
  // removed this for new communication model (soccerserver 8.*)
  /*
  if ((tsnap->ourAction.state == ta) &&
      (tsnap->ourAction.performer == my.number || ta == TA_REACT)) {
    if (time < lastReportTime + 6)
      return;
  }
  */
  tsnap->reportOurTacticalAction(ta, my.number);
  lastReportTime = time;
}


void
BallHandlingModule::findBall()
{
  // turn continuously to find ball on field (quite simple algorithm)
  if( !lookingForBall )
    {   // we just started the search this timestep
      lookingForBall = true;
    
      // turn left or right to find ball ? (take last ball dir into account)
      // NOTE: These values should work for view width NARROW and NORMAL
      turnAngle = ( snap->my.bodyAngleTo( snap->ball ) > 0 ) ?
	DEG_TO_RAD( +40.0 ) : DEG_TO_RAD( -40.0 );

      BHTRC(MLog("find -> started ball search"));
    }

  BHTRC(MLog("find -> continuing ball search"));
  if( time - snap->visualTime <= 1 )
    {
      skills->basicTurn( turnAngle );
      findBallTurnedBody = true;   // turn head eval needs that information
    }
  else
    skills->noAction();
}


void
BallHandlingModule::playerIntercept()
{
  // calculate earliest position where rolling ball can be reached, and run
  // to that position (takes slight changes of this position into account)

  const double goaliePrecision = 0.8 * CATCHABLE_AREA_W;
  const double playerPrecision = skills->kickRange;   // will not work with het. players!

  bool  isGoalie = ( snap->my.number == snap->ourGoalieNo );
  bool  meetInPenaltyArea = true;   // only needed for goalie

  const double secBorder = 2.0;   // safety distance from edge of field
  WorldBall meet;   // first point where ball can be intercepted (with controlBall)
  AdvSkills::TurtleCommand turtleParams;

  // see what controlBall will report
  if( isGoalie )
    {
      turtleParams = skills->controlBall( snap, snap->senseBody.stamina, goaliePrecision, meet );
      if( !meet.inOurPenaltyArea() )
	{
	  turtleParams = skills->controlBall( snap, availableStamina, playerPrecision, meet );
	  meetInPenaltyArea = false;
	}
      // check if ball crosses goal line before controlBall reaches it
      if( snap->ball.pos.getX() > -0.5 * FIELD_LENGTH && meet.pos.getX() < -0.5 * FIELD_LENGTH )
	{
	  // check if ball crosses the line between the goal posts
	  double factor = - ( 0.5 * FIELD_LENGTH + snap->ball.pos.getX() ) /
	    ( meet.pos.getX() - snap->ball.pos.getX() );
	  if( fabs( snap->ball.pos.getY() + factor *
		    ( meet.pos.getY() - snap->ball.pos.getY() ) ) <= 0.5 * GOAL_WIDTH + 1.0 )
	    {
	      // simulate back until position is in field
	      while( meet.pos.getX() < -0.5 * FIELD_LENGTH )
		{
		  meet.v *= 1.0 / BALL_DECAY;
		  meet.pos -= meet.v;
		}
	      // try to get ball just before it enters the goal
	      BHTRC(MLog("Using goToPosition"));
	      // turtleParams = skills->goToPositionTimed( snap, meet.pos, availableStamina, goaliePrecision );
	      turtleParams = skills->goToPosition( snap, meet.pos, goaliePrecision, 2 * MAXPOWER,
						   true, snap->my.bodyDir );
	      skills->exec( snap, turtleParams );
	      return;
	    }
	}
    }
  else
    turtleParams = skills->controlBall( snap, availableStamina, playerPrecision, meet );


  BHTRC(MLog("Using controlBall"));

  if (turtleParams.state != AdvSkills::TFINISHED) {
    // not at the ball yet
    if (turtleParams.time > 2) {
      // We will need some time (more than 2 timesteps) until we get the ball.
      // check 2 things: 1. is the minimum intercept stable?
      // 2. if not, can we intercept later ?
      // this is not done and will not be done in the skills because
      // it heavily depends on other players
      int closestTime = turtleParams.time;
      int timeToReach = (tsnap->firstDangerousOpponent == UNKNOWN) ? 20 :
	tsnap->otherTeamReachBallTime(tsnap->firstDangerousOpponent);

      // [!] maybe it is a good idea to always relax at least a little bit
      if (timeToReach <= closestTime)
	timeToReach = closestTime + 1;

      // now try letting the ball roll.
      // accept if
      // 1. the difference between the reachTimes is large
      //    this means an instable intercept
      // 2. we significantly gain space
      // problems:
      // it is likely that the closestTime is MUCH smaller
      // than the next possible time. This is the instable case
      // if possible, the next valid meetpos should be used

      // we actually decide the stability by the previous point
      bool prevStable = true;
      AdvSkills::TurtleCommand prevCmd = turtleParams;
      int simTime = closestTime;
      WorldBall simBall = meet;

#ifdef BALLHTRACE
      ostrstream tmp;
      tmp << "sim -> ttr=" << timeToReach << " st=" << simTime
	  << (char) 0;
      char * str = tmp.str();
      BHTRC(MLog(str));
      delete str;
#endif

      bool cont = true;
      while (cont && (timeToReach >= simTime)) {
	simBall.simulate();
	simTime++;
	// only use simulated position if ball safely inside field 
	// and not too far in the front (why that??? cm)
	if (!(simBall.pos.getX() > -FIELD_LENGTH/2 + secBorder) ||
	    !(simBall.pos.getX() < FIELD_LENGTH/5) ||
	    !(fabs(simBall.pos.getY()) < FIELD_WIDTH/2 - secBorder))
	  {
	    BHTRC(MLog("PlayerIntercept simBall outside field or too far"));
	    break;
	  };
	AdvSkills::TurtleCommand tmp;
	if( isGoalie )
	  {
	    if( simBall.inOurPenaltyArea() )
	      tmp = skills->goToPositionTimed( snap, simBall.pos, availableStamina, goaliePrecision );
	    else if( meetInPenaltyArea )
	      {
		// if first meet was in penalty area stop simulation when simBall leaves it
		cont = false;
		continue;
	      }
	    else
	      tmp = skills->goToPositionTimed( snap, simBall.pos, availableStamina, playerPrecision );
	  }
	else
	  tmp = skills->goToPositionTimed( snap, simBall.pos, availableStamina, playerPrecision );

	if (tmp.time > simTime) {
	  // not stable
	  prevStable = false;
	  BHTRC(MLog("sim -> instability detected"));
	}
	else if (prevStable) {
	  // we are only interested in stable situations
	  // if we cannot reach one, the loop will keep
	  // the (maybe unsafe) first guess
	  // CM: replaced constant 0.2 by 0.5 (0.2 * BALL_SPEED_MAX is reached after 25 steps
	  // (when ball speed is at max), 0.5 * BALL_SPEED_MAX after 10 steps)
	  // REMARK: ball speed at "meet" is even higher!
	  if ((meet.v.getX() < 0.5 * BALL_SPEED_MAX) ||
	      (timeToReach == simTime)) {
	    // stable situation, but it doesn't make sense
	    // to let the ball roll or the simulation ends
	    turtleParams = prevCmd;
	    BHTRC(MLog("sim -> found better solution"));
	    cont = false;
	  }
	  prevStable = true;
	  prevCmd    = tmp;
	}
      }
    }

    skills->exec(snap, turtleParams);
   
 
#ifdef BALLHTRACE
    ostrstream tmp;
    tmp << "intercept -> needed time = " << turtleParams.time << (char) 0;
    char * str = tmp.str();
    BHTRC(MLog(str));
    delete str;
#endif
    return;
  }

  // finished - turn to ball
  AdvSkills::TurnCommand turnParams =
    skills->turn(snap,(snap->ball.pos - snap->my.pos).getAngle());
  skills->exec(snap, turnParams);
  BHTRC (MLog("turning to ball"));
}


void
BallHandlingModule::fastdribble()
{
  Vector2d neededBallVel; // speed the ball will initially need for that fast dribble

  neededBallVel.setRadAng( ( 1 - BALL_DECAY ) / ( 1 - power( BALL_DECAY, fd_numSteps ) ) * fd_dist, dribbleDir );
  
#ifdef BALLHTRACE
  ostrstream tmp;
  tmp << "fastdribble " << dribbleDir << " " << fd_numSteps << " " << fd_dashPower << (char)0;
  char * str = tmp.str();
  BHTRC(MLog(str));
  delete str;
#endif

  AdvSkills::ExtKickCommand kickParams = skills->kick( snap, neededBallVel );
    
  if( kickParams.state == AdvSkills::KRANGE )
    skills->noAction();
  else
    skills->exec( snap, kickParams );
}


void
BallHandlingModule::dribble() {
  // keep ball under control while running
#ifdef BALLHTRACE
  ostrstream tmp;
  tmp << "dribble d=" << dribbleDir << " h=" << holdDir << (char)0;
  char * str = tmp.str();
  BHTRC(MLog(str));
  delete str;
#endif
  AdvSkills::DribbleCommand dribbleParams =
    skills->dribble( snap, dribbleDir, holdDir, dribbleStamina );
  skills->exec( snap, dribbleParams );
  BHTRC( LINE ( my.pos, my.pos + radAng2Vector2d ( 10.0, dribbleDir ) ) );
}


void
BallHandlingModule::place() {
  // move ball to placeDir on a safe circle around us.

  if (tsnap);
  
  Vector2d placePos;
  placePos.setRadAng( skills->safeCircle, placeDir );
  
  AdvSkills::KickCommand kickParams = skills->placeBall( snap, placePos );
  if (kickParams.state == AdvSkills::KPLACING) {

    BHTRC(LINE(snap->my.pos, radAng2Vector2d(1, placeDir)));
    skills->exec( snap, kickParams);
  } else {
    BHTRC(MLog("place -> trying to place ball, but failed!"));
    skills->noAction();
  }
}


void
BallHandlingModule::goalieCatch()
{
  // try to catch ball directly
  skills->exec(snap, catchParams); // params computed in evaluate()
}

